/*
 * Implementation of the proxy that uses the best available Rhythmbox proxy.
 *
 * Rhythmbox Applet
 * Copyright (C) 2004-2005 Paul Kuliniewicz <kuliniew@purdue.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.
 */

#include <config.h>

#include "rb-any-proxy.h"
#include "rb-bonobo-proxy.h"
#ifdef WITH_DBUS
#  include "rb-dbus-proxy.h"
#endif

#include <string.h>

struct _RbAnyProxyPrivate
{
	RbProxy *bonobo;
	RbProxy *dbus;

	RbProxy *active;
};

typedef enum
{
	PROP_NONE,
	PROP_SONG,
	PROP_ELAPSED,
	PROP_RATING,
	PROP_CONNECTED,
	PROP_PLAYING,
	PROP_PAUSABLE,
} RbAnyProxyProperty;

static GObjectClass *parent_class;

/* Function declarations */

static void proxy_interface_init (gpointer g_iface, gpointer data);

static void rb_any_proxy_class_init (RbAnyProxyClass *klass);
static void rb_any_proxy_init (RbAnyProxy *dproxy);

static void rb_any_proxy_dispose (GObject *object);
static void rb_any_proxy_finalize (GObject *object);

static void rb_any_proxy_get_property (GObject *object,
					guint prop_id,
					GValue *value,
					GParamSpec *pspec);
static void rb_any_proxy_set_property (GObject *object,
					guint prop_id,
					const GValue *value,
					GParamSpec *pspec);

static void rb_any_proxy_play_pause_stop (RbProxy *proxy);
static void rb_any_proxy_previous (RbProxy *proxy);
static void rb_any_proxy_next (RbProxy *proxy);

static void create_bonobo_proxy (RbAnyProxy *aproxy);
static void destroy_bonobo_proxy (RbAnyProxy *aproxy);

static void create_dbus_proxy (RbAnyProxy *aproxy);
static void destroy_dbus_proxy (RbAnyProxy *aproxy);

static void choose_active_proxy (RbAnyProxy *aproxy);

static void notify_cb (RbProxy *proxy, GParamSpec *pspec, RbAnyProxy *aproxy);

/* GType stuff */

GType
rb_any_proxy_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (RbAnyProxyClass),			/* class_size */
			NULL,						/* base_init */
			NULL,						/* base_finalize */
			(GClassInitFunc) rb_any_proxy_class_init,	/* class_init */
			NULL,						/* class_finalize */
			NULL,						/* class_data */
			sizeof (RbAnyProxy),				/* instance_size */
			0,						/* n_preallocs */
			(GInstanceInitFunc) rb_any_proxy_init,		/* instance_init */
			NULL
		};

		static const GInterfaceInfo proxy_info = {
			(GInterfaceInitFunc) proxy_interface_init,	/* interface_init */
			NULL,						/* interface_finalize */
			NULL						/* interface_data */
		};

		type = g_type_register_static (G_TYPE_OBJECT, "RbAnyProxy", &info, 0);
		g_type_add_interface_static (type, RB_TYPE_PROXY, &proxy_info);
	}

	return type;
}

static void
proxy_interface_init (gpointer g_iface, gpointer data)
{
	RbProxyClass *klass = (RbProxyClass *) g_iface;

	klass->play_pause_stop = rb_any_proxy_play_pause_stop;
	klass->previous = rb_any_proxy_previous;
	klass->next = rb_any_proxy_next;
}

static void
rb_any_proxy_class_init (RbAnyProxyClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;
	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = rb_any_proxy_dispose;
	object_class->finalize = rb_any_proxy_finalize;
	object_class->get_property = rb_any_proxy_get_property;
	object_class->set_property = rb_any_proxy_set_property;

	g_object_class_override_property (object_class, PROP_SONG, "song");
	g_object_class_override_property (object_class, PROP_ELAPSED, "elapsed");
	g_object_class_override_property (object_class, PROP_RATING, "rating");
	g_object_class_override_property (object_class, PROP_CONNECTED, "connected");
	g_object_class_override_property (object_class, PROP_PLAYING, "playing");
	g_object_class_override_property (object_class, PROP_PAUSABLE, "pausable");
}

static void
rb_any_proxy_init (RbAnyProxy *aproxy)
{
	aproxy->priv = g_new0 (RbAnyProxyPrivate, 1);

	aproxy->priv->bonobo = NULL;
	aproxy->priv->dbus = NULL;

	aproxy->priv->active = NULL;
}

static void
rb_any_proxy_dispose (GObject *object)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (object);

	aproxy->priv->active = NULL;

	if (aproxy->priv->bonobo != NULL)
		destroy_bonobo_proxy (aproxy);

	if (aproxy->priv->dbus != NULL)
		destroy_dbus_proxy (aproxy);

	parent_class->dispose (object);
}

static void
rb_any_proxy_finalize (GObject *object)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (object);

	g_free (aproxy->priv);

	parent_class->finalize (object);
}

/* Public interface */

RbProxy *
rb_any_proxy_new (void)
{
	RbAnyProxy *aproxy;

	aproxy = g_object_new (RB_TYPE_ANY_PROXY, NULL);

	create_bonobo_proxy (aproxy);
	create_dbus_proxy (aproxy);

	choose_active_proxy (aproxy);

	return RB_PROXY (aproxy);
}

/* GObject overrides */

static void
rb_any_proxy_get_property (GObject *object,
			   guint prop_id,
			   GValue *value,
			   GParamSpec *pspec)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (object);

	g_return_if_fail (aproxy != NULL);

	if (aproxy->priv->active != NULL)
	{
		g_object_get_property (G_OBJECT (aproxy->priv->active),
				       pspec->name,
				       value);
	}
	else if (prop_id == PROP_CONNECTED)
		g_value_set_boolean (value, FALSE);
	else
		g_param_value_set_default (pspec, value);
}

static void
rb_any_proxy_set_property (GObject *object,
			   guint prop_id,
			   const GValue *value,
			   GParamSpec *pspec)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (object);

	g_return_if_fail (aproxy != NULL);

	if (aproxy->priv->active != NULL)
	{
		g_object_set_property (G_OBJECT (aproxy->priv->active),
				       pspec->name,
				       value);
	}
}

/* RbProxy overrides */

static void
rb_any_proxy_play_pause_stop (RbProxy *proxy)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (proxy);

	g_return_if_fail (aproxy != NULL);

	if (aproxy->priv->active != NULL)
		rb_proxy_play_pause_stop (aproxy->priv->active);
}

static void
rb_any_proxy_previous (RbProxy *proxy)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (proxy);

	g_return_if_fail (aproxy != NULL);

	if (aproxy->priv->active != NULL)
		rb_proxy_previous (aproxy->priv->active);
}

static void
rb_any_proxy_next (RbProxy *proxy)
{
	RbAnyProxy *aproxy = RB_ANY_PROXY (proxy);

	g_return_if_fail (aproxy != NULL);

	if (aproxy->priv->active != NULL)
		rb_proxy_next (aproxy->priv->active);
}

/* Internal functions */

static void
create_bonobo_proxy (RbAnyProxy *aproxy)
{
	g_return_if_fail (aproxy->priv->bonobo == NULL);

	aproxy->priv->bonobo = rb_bonobo_proxy_new ();
	g_signal_connect (aproxy->priv->bonobo, "notify",
			  G_CALLBACK (notify_cb), aproxy);
}

static void
destroy_bonobo_proxy (RbAnyProxy *aproxy)
{
	g_return_if_fail (aproxy->priv->bonobo != NULL);

	/* Prevent a nasty feedback if notify::connected gets triggered. */

	g_signal_handlers_disconnect_by_func (aproxy->priv->bonobo, notify_cb, aproxy);

	g_object_unref (aproxy->priv->bonobo);
	aproxy->priv->bonobo = NULL;
}

static void
create_dbus_proxy (RbAnyProxy *aproxy)
{
#ifdef WITH_DBUS
	g_return_if_fail (aproxy->priv->dbus == NULL);

	aproxy->priv->dbus = rb_dbus_proxy_new ();
	g_signal_connect (aproxy->priv->dbus, "notify",
			  G_CALLBACK (notify_cb), aproxy);
#endif
}

static void
destroy_dbus_proxy (RbAnyProxy *aproxy)
{
#ifdef WITH_DBUS
	g_return_if_fail (aproxy->priv->dbus != NULL);

	/* Prevent a nasty feedback if notify::connected gets triggered. */

	g_signal_handlers_disconnect_by_func (aproxy->priv->dbus, notify_cb, aproxy);

	g_object_unref (aproxy->priv->dbus);
	aproxy->priv->dbus = NULL;
#endif
}

static void
choose_active_proxy (RbAnyProxy *aproxy)
{
	/*
	 * Policy: Prefer D-Bus, but fall back to Bonobo if necessary.
	 *
	 * Since the Bonobo interface inherently requires frequent polling,
	 * destroy that proxy whenever a connection to Rhythmbox via D-Bus
	 * exists, to reduce needless overhead for both us and Rhythmbox.
	 * (Bonobo isn't exactly a speed demon....)
	 */

	gboolean connected = FALSE;

	if (aproxy->priv->dbus != NULL)
		g_object_get (aproxy->priv->dbus, "connected", &connected, NULL);
	if (connected)
	{
		aproxy->priv->active = aproxy->priv->dbus;
		if (aproxy->priv->bonobo != NULL)
			destroy_bonobo_proxy (aproxy);
		return;
	}

	if (aproxy->priv->bonobo == NULL)
		create_bonobo_proxy (aproxy);

	g_object_get (aproxy->priv->bonobo, "connected", &connected, NULL);
	if (connected)
	{
		aproxy->priv->active = aproxy->priv->bonobo;
		return;
	}

	aproxy->priv->active = NULL;
}

/* Callbacks */

static void
notify_cb (RbProxy *proxy, GParamSpec *pspec, RbAnyProxy *aproxy)
{
	if (strcmp(pspec->name, "connected") == 0)
	{
		choose_active_proxy (aproxy);
		g_object_notify (G_OBJECT (aproxy), pspec->name);
	}
	else if (proxy == aproxy->priv->active)
		g_object_notify (G_OBJECT (aproxy), pspec->name);
}
