/*
 * Implementation of the proxy around Rhythmbox's Bonobo interface.
 *
 * Rhythmbox Applet
 * Copyright (C) 2004-2006 Paul Kuliniewicz <kuliniew@purdue.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.
 */

/* FIXME: Never call rb_bonobo_proxy_* functions directly, to allow for
 *        (theoretical) overriding by further derived classes.  Or no?
 */

#include <config.h>

#include "rb-bonobo-proxy.h"

#include <libbonobo.h>

#include "Rhythmbox.h"
#include "rb-song-metadata.h"

#define POLL_DELAY		5000
#define STATUS_DELAY		500

struct _RbBonoboProxyPrivate
{
	GNOME_Rhythmbox rb;
	Bonobo_PropertyBag pb;
	CORBA_Environment ev;

	Bonobo_Listener song_listener;
	Bonobo_Listener playing_listener;

	RbSongMetadata *song;
	gdouble rating;

	guint poll;
	guint timer;
};

typedef enum
{
	PROP_NONE,
	PROP_SONG,
	PROP_ELAPSED,
	PROP_RATING,
	PROP_CONNECTED,
	PROP_PLAYING,
	PROP_PAUSABLE,
} RbBonoboProxyProperty;

static GObjectClass *parent_class;

/* Function declarations */

static void proxy_interface_init (gpointer g_iface, gpointer data);

static void rb_bonobo_proxy_class_init (RbBonoboProxyClass *klass);
static void rb_bonobo_proxy_init (RbBonoboProxy *bproxy);

static void rb_bonobo_proxy_dispose (GObject *object);
static void rb_bonobo_proxy_finalize (GObject *object);
static void rb_bonobo_proxy_get_property (GObject *object,
					  guint prop_id,
					  GValue *value,
					  GParamSpec *pspec);
static void rb_bonobo_proxy_set_property (GObject *object,
					  guint prop_id,
					  const GValue *value,
					  GParamSpec *pspec);

static void rb_bonobo_proxy_connect (RbProxy *proxy);
static void rb_bonobo_proxy_disconnect (RbProxy *proxy);

static void rb_bonobo_proxy_play_pause_stop (RbProxy *proxy);
static void rb_bonobo_proxy_previous (RbProxy *proxy);
static void rb_bonobo_proxy_next (RbProxy *proxy);

static gboolean is_connected (RbBonoboProxy *bproxy);
static gboolean is_playing (RbBonoboProxy *bproxy);
static gboolean is_pausable (RbBonoboProxy *bproxy);

static void rate_current_song (RbBonoboProxy *bproxy, gdouble rating);

static glong get_play_time (RbBonoboProxy *bproxy);
static void get_song_info (RbBonoboProxy *bproxy);
static void cache_song_metadata (RbBonoboProxy *bproxy, const GNOME_Rhythmbox_SongInfo *info);
static void report_error (RbBonoboProxy *bproxy);

static gboolean poll_cb (RbProxy *proxy);
static void song_change_cb (BonoboListener *listener,
			    const char *event_name,
			    const CORBA_any *any,
			    CORBA_Environment *ev,
			    RbBonoboProxy *bproxy);
static void playing_change_cb (BonoboListener *listener,
			       const char *event_name,
			       const CORBA_any *any,
			       CORBA_Environment *ev,
			       RbBonoboProxy *bproxy);
static gboolean time_update_cb (RbBonoboProxy *bproxy);

/* GType stuff */

GType
rb_bonobo_proxy_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (RbBonoboProxyClass),			/* class_size */
			NULL,						/* base_init */
			NULL,						/* base_finalize */
			(GClassInitFunc) rb_bonobo_proxy_class_init,	/* class_init */
			NULL,						/* class_finalize */
			NULL,						/* class_data */
			sizeof (RbBonoboProxy), 			/* instance_size */
			0,						/* n_preallocs */
			(GInstanceInitFunc) rb_bonobo_proxy_init,	/* instance_init */
			NULL
		};

		static const GInterfaceInfo proxy_info = {
			(GInterfaceInitFunc) proxy_interface_init,	/* interface_init */
			NULL,						/* interface_finalize */
			NULL						/* interface_data */
		};

		type = g_type_register_static (G_TYPE_OBJECT, "RbBonoboProxy", &info, 0);
		g_type_add_interface_static (type, RB_TYPE_PROXY, &proxy_info);
	}

	return type;
}

static void
proxy_interface_init (gpointer g_iface, gpointer data)
{
	RbProxyClass *klass = (RbProxyClass *) g_iface;

	klass->play_pause_stop = rb_bonobo_proxy_play_pause_stop;
	klass->previous = rb_bonobo_proxy_previous;
	klass->next = rb_bonobo_proxy_next;
}

static void
rb_bonobo_proxy_class_init (RbBonoboProxyClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;
	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = rb_bonobo_proxy_dispose;
	object_class->finalize = rb_bonobo_proxy_finalize;
	object_class->get_property = rb_bonobo_proxy_get_property;
	object_class->set_property = rb_bonobo_proxy_set_property;

	g_object_class_override_property (object_class, PROP_SONG, "song");
	g_object_class_override_property (object_class, PROP_ELAPSED, "elapsed");
	g_object_class_override_property (object_class, PROP_RATING, "rating");
	g_object_class_override_property (object_class, PROP_CONNECTED, "connected");
	g_object_class_override_property (object_class, PROP_PLAYING, "playing");
	g_object_class_override_property (object_class, PROP_PAUSABLE, "pausable");
}

static void
rb_bonobo_proxy_init (RbBonoboProxy *bproxy)
{
	bproxy->priv = g_new0 (RbBonoboProxyPrivate, 1);

	if (!bonobo_is_initialized ())
	{
		gint fake_argc = 0;
		bonobo_init (&fake_argc, NULL);
		bonobo_activate ();
	}

	bproxy->priv->rb = CORBA_OBJECT_NIL;
	bproxy->priv->pb = CORBA_OBJECT_NIL;

	bproxy->priv->song_listener = CORBA_OBJECT_NIL;
	bproxy->priv->playing_listener = CORBA_OBJECT_NIL;

	CORBA_exception_init (&bproxy->priv->ev);

	bproxy->priv->song = NULL;
	bproxy->priv->rating = 0.0;
}

static void
rb_bonobo_proxy_dispose (GObject *object)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (object);

	rb_bonobo_proxy_disconnect (RB_PROXY (bproxy));

	/* Disconnect starts the poll timer, so need to get rid of it. */

	g_source_remove (bproxy->priv->poll);
	bproxy->priv->poll = 0;

	parent_class->dispose (object);
}

static void
rb_bonobo_proxy_finalize (GObject *object)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (object);

	CORBA_exception_free (&bproxy->priv->ev);
	g_free (bproxy->priv);

	parent_class->finalize (object);
}

/* Public interface */

RbProxy *
rb_bonobo_proxy_new (void)
{
	RbBonoboProxy *bproxy;

	bproxy = g_object_new (RB_TYPE_BONOBO_PROXY, NULL);
	bproxy->priv->poll = 0;
	bproxy->priv->timer = 0;
	rb_bonobo_proxy_connect (RB_PROXY (bproxy));

	return RB_PROXY (bproxy);
}

/* GObject overrides */

static void
rb_bonobo_proxy_get_property (GObject *object,
			      guint prop_id,
			      GValue *value,
			      GParamSpec *pspec)
{
	RbBonoboProxy *bproxy;

	g_return_if_fail (object != NULL);

	bproxy = RB_BONOBO_PROXY (object);

	switch (prop_id)
	{
	case PROP_SONG:
		g_value_set_object (value, bproxy->priv->song);
		break;
	case PROP_ELAPSED:
		g_value_set_long (value, get_play_time (bproxy));
		break;
	case PROP_RATING:
		g_value_set_double (value, bproxy->priv->rating);
		break;
	case PROP_CONNECTED:
		g_value_set_boolean (value, is_connected (bproxy));
		break;
	case PROP_PLAYING:
		g_value_set_boolean (value, is_playing (bproxy));
		break;
	case PROP_PAUSABLE:
		g_value_set_boolean (value, is_pausable (bproxy));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
rb_bonobo_proxy_set_property (GObject *object,
			      guint prop_id,
			      const GValue *value,
			      GParamSpec *pspec)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (object);

	g_return_if_fail (bproxy != NULL);

	switch (prop_id)
	{
	case PROP_RATING:
		rate_current_song (bproxy, g_value_get_double (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

/* RbProxy overrides */

static void
rb_bonobo_proxy_connect (RbProxy *proxy)
{
	RbBonoboProxy *bproxy;
	GClosure *closure;

	g_return_if_fail (proxy != NULL);

	bproxy = RB_BONOBO_PROXY (proxy);

	if (is_connected (bproxy))
		return;

	if (bproxy->priv->poll != 0)
	{
		g_source_remove (bproxy->priv->poll);
		bproxy->priv->poll = 0;
	}

	bproxy->priv->rb = bonobo_activation_activate (
		"((repo_ids.has('IDL:GNOME/Rhythmbox:1.0')) AND (_active == TRUE))",
		NULL, 0, NULL, &bproxy->priv->ev);
	if (bproxy->priv->rb == CORBA_OBJECT_NIL)
	{
		bproxy->priv->poll = g_timeout_add (
			POLL_DELAY, (GSourceFunc) poll_cb, bproxy);
		return;
	}

	bproxy->priv->pb = GNOME_Rhythmbox_getPlayerProperties (bproxy->priv->rb, &bproxy->priv->ev);
	if (BONOBO_EX (&bproxy->priv->ev))
		report_error (bproxy);

	/*
	 * If bonobo_event_source_client_add_listener gave some way to remove
	 * a listener, we'd use that.  But since it doesn't, we get to set up
	 * event listeners the (relatively) hard way.
	 */

	closure = g_cclosure_new (G_CALLBACK (song_change_cb), bproxy, NULL);
	bproxy->priv->song_listener = bonobo_event_source_client_add_listener_full (bproxy->priv->pb,
										    closure,
										    "Bonobo/Property:change:song",
										    &bproxy->priv->ev);
	if (BONOBO_EX (&bproxy->priv->ev))
		report_error (bproxy);

	closure = g_cclosure_new (G_CALLBACK (playing_change_cb), bproxy, NULL);
	bproxy->priv->playing_listener = bonobo_event_source_client_add_listener_full (bproxy->priv->pb,
										       closure,
										       "Bonobo/Property:change:playing",
										       &bproxy->priv->ev);
	if (BONOBO_EX (&bproxy->priv->ev))
		report_error (bproxy);

	if (is_connected (bproxy))
	{
		bproxy->priv->timer = g_timeout_add (
			STATUS_DELAY, (GSourceFunc) time_update_cb, bproxy);
		g_object_notify (G_OBJECT (bproxy), "connected");
		get_song_info (bproxy);
	}
}

static void
rb_bonobo_proxy_disconnect (RbProxy *proxy)
{
	RbBonoboProxy *bproxy;

	g_return_if_fail (proxy != NULL);

	bproxy = RB_BONOBO_PROXY (proxy);

	if (!is_connected (bproxy))
		return;

	if (bproxy->priv->timer != 0)
	{
		g_source_remove (bproxy->priv->timer);
		bproxy->priv->timer = 0;
	}

	if (bproxy->priv->song_listener != CORBA_OBJECT_NIL)
	{
		bonobo_event_source_client_remove_listener (bproxy->priv->pb,
							    bproxy->priv->song_listener,
							    NULL);
		bproxy->priv->song_listener = CORBA_OBJECT_NIL;
	}

	if (bproxy->priv->playing_listener != CORBA_OBJECT_NIL)
	{
		bonobo_event_source_client_remove_listener (bproxy->priv->pb,
							    bproxy->priv->playing_listener,
							    NULL);
		bproxy->priv->playing_listener = CORBA_OBJECT_NIL;
	}

	if (bproxy->priv->pb != CORBA_OBJECT_NIL)
	{
		bonobo_object_release_unref ((Bonobo_Unknown) bproxy->priv->pb,
					     NULL);
		bproxy->priv->pb = CORBA_OBJECT_NIL;
	}

	if (bproxy->priv->rb != CORBA_OBJECT_NIL)
		bproxy->priv->rb = CORBA_OBJECT_NIL;

	g_object_notify (G_OBJECT (proxy), "song");
	g_object_notify (G_OBJECT (proxy), "elapsed");
	g_object_notify (G_OBJECT (proxy), "playing");
	g_object_notify (G_OBJECT (proxy), "connected");

	bproxy->priv->poll = g_timeout_add (POLL_DELAY, (GSourceFunc) poll_cb, bproxy);
}

static void
rb_bonobo_proxy_play_pause_stop (RbProxy *proxy)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (proxy);

	g_return_if_fail (bproxy != NULL);

	rb_bonobo_proxy_connect (proxy);
	if (is_connected (bproxy))
	{
		GNOME_Rhythmbox_playPause (bproxy->priv->rb, &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
	}
}

static void
rb_bonobo_proxy_previous (RbProxy *proxy)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (proxy);

	g_return_if_fail (bproxy != NULL);

	if (is_connected (bproxy))
	{
		GNOME_Rhythmbox_previous (bproxy->priv->rb, &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
	}
}

static void
rb_bonobo_proxy_next (RbProxy *proxy)
{
	RbBonoboProxy *bproxy = RB_BONOBO_PROXY (proxy);

	g_return_if_fail (bproxy != NULL);

	if (is_connected (bproxy))
	{
		GNOME_Rhythmbox_next (bproxy->priv->rb, &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
	}
}

/* Private functions */

static gboolean
is_connected (RbBonoboProxy *bproxy)
{
	return (bproxy->priv->rb != CORBA_OBJECT_NIL);
}

static gboolean
is_playing (RbBonoboProxy *bproxy)
{
	gboolean playing = FALSE;

	if (is_connected (bproxy))
	{
		playing = bonobo_pbclient_get_boolean (bproxy->priv->pb,
						       "playing",
						       &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
	}

	return playing;
}

static gboolean
is_pausable (RbBonoboProxy *bproxy)
{
	if (bproxy->priv->song != NULL)
	{
		glong duration;
		g_object_get (bproxy->priv->song, "duration", &duration, NULL);
		return (duration > 0);
	}
	else
		return FALSE;
}

static void
rate_current_song (RbBonoboProxy *bproxy, gdouble rating)
{
	g_return_if_fail (rating >= 0.0);
	g_return_if_fail (rating <= 5.0);

	if (is_connected (bproxy))
	{
		GNOME_Rhythmbox_setRating (bproxy->priv->rb, rating, &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);

		bproxy->priv->rating = rating;
		g_object_notify (G_OBJECT (bproxy), "rating");
	}
}

static glong
get_play_time (RbBonoboProxy *bproxy)
{
	glong play_time = -1;

	if (is_connected (bproxy))
	{
		play_time = GNOME_Rhythmbox_getPlayingTime (bproxy->priv->rb,
							    &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
	}

	return play_time;
}

static void
get_song_info (RbBonoboProxy *bproxy)
{
	const GNOME_Rhythmbox_SongInfo *info = NULL;

	g_return_if_fail (bproxy != NULL);

	if (is_connected (bproxy))
	{
		CORBA_any *any = bonobo_pbclient_get_value (bproxy->priv->pb,
							    "song",
							    TC_GNOME_Rhythmbox_SongInfo,
							    &bproxy->priv->ev);
		if (BONOBO_EX (&bproxy->priv->ev))
			report_error (bproxy);
		if (any != NULL)
			info = (const GNOME_Rhythmbox_SongInfo *) any->_value;
	}

	cache_song_metadata (bproxy, info);
}

static void
cache_song_metadata (RbBonoboProxy *bproxy, const GNOME_Rhythmbox_SongInfo *info)
{
	if (bproxy->priv->song != NULL)
		g_object_unref (bproxy->priv->song);

	if (info != NULL)
	{
		bproxy->priv->song = rb_song_metadata_new (info->title,
							   info->artist,
							   info->album,
							   info->duration);
		bproxy->priv->rating = (gdouble) info->rating;
	}
	else
	{
		bproxy->priv->song = NULL;
		bproxy->priv->rating = 0.0;
	}

	g_object_notify (G_OBJECT (bproxy), "song");
	g_object_notify (G_OBJECT (bproxy), "rating");
}

static void
report_error (RbBonoboProxy *bproxy)
{
	gchar *what = bonobo_exception_get_text (&bproxy->priv->ev);
	g_warning ("Bonobo error: %s", what);
	g_free (what);

	rb_bonobo_proxy_disconnect (RB_PROXY (bproxy));
}

/* Callbacks */

static gboolean
poll_cb (RbProxy *proxy)
{
	rb_bonobo_proxy_connect (proxy);
	return !is_connected (RB_BONOBO_PROXY (proxy));
}

static void
song_change_cb (BonoboListener *listener,
		const char *event_name,
		const CORBA_any *any,
		CORBA_Environment *ev,
		RbBonoboProxy *bproxy)
{
	const GNOME_Rhythmbox_SongInfo *info;

	g_return_if_fail (CORBA_TypeCode_equivalent (any->_type,
						     TC_GNOME_Rhythmbox_SongInfo,
						     NULL));
	info = (const GNOME_Rhythmbox_SongInfo *) any->_value;

	cache_song_metadata (bproxy, info);
}

static void
playing_change_cb (BonoboListener *listener,
		   const char *event_name,
		   const CORBA_any *any,
		   CORBA_Environment *ev,
		   RbBonoboProxy *bproxy)
{
	glong elapsed;

	g_object_get (bproxy, "elapsed", &elapsed, NULL);
	if (elapsed < 0)
		cache_song_metadata (bproxy, NULL);

	g_object_notify (G_OBJECT (bproxy), "playing");
}

static gboolean
time_update_cb (RbBonoboProxy *bproxy)
{
	g_object_notify (G_OBJECT (bproxy), "elapsed");

	return is_connected (bproxy);
}
