/*
 * Implementation of the proxy around Rhythmbox's D-Bus interface.
 *
 * Rhythmbox Applet
 * Copyright (C) 2004-2006 Paul Kuliniewicz <kuliniew@purdue.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.
 */

#include <config.h>

#include "rb-dbus-proxy.h"
#include "rb-song-metadata.h"

#include <dbus/dbus-glib.h>
#include <glib/ghash.h>
#include <string.h>

struct _RbDBusProxyPrivate
{
	DBusGConnection *connection;
	DBusGProxy *bus_proxy;
	DBusGProxy *player_proxy;
	DBusGProxy *shell_proxy;

	RbSongMetadata *song;
	gdouble rating;

	gboolean connected;
	gboolean playing;
	gboolean pausable;
	gchar *uri;
	guint elapsed;
};

typedef enum
{
	PROP_NONE,
	PROP_SONG,
	PROP_ELAPSED,
	PROP_RATING,
	PROP_CONNECTED,
	PROP_PLAYING,
	PROP_PAUSABLE,
} RbDBusProxyProperty;

static GObjectClass *parent_class;

/* Function declarations */

static void proxy_interface_init (gpointer g_iface, gpointer data);

static void rb_dbus_proxy_class_init (RbDBusProxyClass *klass);
static void rb_dbus_proxy_init (RbDBusProxy *dproxy);

static void rb_dbus_proxy_dispose (GObject *object);
static void rb_dbus_proxy_finalize (GObject *object);
static void rb_dbus_proxy_get_property (GObject *object,
					guint prop_id,
					GValue *value,
					GParamSpec *pspec);
static void rb_dbus_proxy_set_property (GObject *object,
					guint prop_id,
					const GValue *value,
					GParamSpec *pspec);

static void rb_dbus_proxy_play_pause_stop (RbProxy *proxy);
static void rb_dbus_proxy_previous (RbProxy *proxy);
static void rb_dbus_proxy_next (RbProxy *proxy);

static void rate_current_song (RbDBusProxy *dproxy, gdouble rating);

static void refresh (RbDBusProxy *dproxy);
static void refresh_playing (RbDBusProxy *dproxy);
static void refresh_elapsed (RbDBusProxy *dproxy);
static void refresh_playing_uri (RbDBusProxy *dproxy);
static void cache_song_metadata (RbDBusProxy *dproxy);

static void name_owner_changed_cb (DBusGProxy *bus_proxy,
				   const gchar *name,
				   const gchar *old_owner,
				   const gchar *new_owner,
				   RbDBusProxy *dproxy);

static void playing_changed_cb (DBusGProxy *player_proxy,
				gboolean playing,
				RbDBusProxy *dproxy);
static void uri_changed_cb (DBusGProxy *player_proxy,
			    const gchar *uri,
			    RbDBusProxy *dproxy);
static void elapsed_changed_cb (DBusGProxy *player_proxy,
				guint elapsed,
				RbDBusProxy *dproxy);

/* GType stuff */

GType
rb_dbus_proxy_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (RbDBusProxyClass),			/* class_size */
			NULL,						/* base_init */
			NULL,						/* base_finalize */
			(GClassInitFunc) rb_dbus_proxy_class_init,	/* class_init */
			NULL,						/* class_finalize */
			NULL,						/* class_data */
			sizeof (RbDBusProxy),				/* instance_size */
			0,						/* n_preallocs */
			(GInstanceInitFunc) rb_dbus_proxy_init,		/* instance_init */
			NULL
		};

		static const GInterfaceInfo proxy_info = {
			(GInterfaceInitFunc) proxy_interface_init,	/* interface_init */
			NULL,						/* interface_finalize */
			NULL						/* interface_data */
		};

		type = g_type_register_static (G_TYPE_OBJECT, "RbDBusProxy", &info, 0);
		g_type_add_interface_static (type, RB_TYPE_PROXY, &proxy_info);
	}

	return type;
}

static void
proxy_interface_init (gpointer g_iface, gpointer data)
{
	RbProxyClass *klass = (RbProxyClass *) g_iface;

	klass->play_pause_stop = rb_dbus_proxy_play_pause_stop;
	klass->previous = rb_dbus_proxy_previous;
	klass->next = rb_dbus_proxy_next;
}

static void
rb_dbus_proxy_class_init (RbDBusProxyClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;
	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = rb_dbus_proxy_dispose;
	object_class->finalize = rb_dbus_proxy_finalize;
	object_class->get_property = rb_dbus_proxy_get_property;
	object_class->set_property = rb_dbus_proxy_set_property;

	g_object_class_override_property (object_class, PROP_SONG, "song");
	g_object_class_override_property (object_class, PROP_ELAPSED, "elapsed");
	g_object_class_override_property (object_class, PROP_RATING, "rating");
	g_object_class_override_property (object_class, PROP_CONNECTED, "connected");
	g_object_class_override_property (object_class, PROP_PLAYING, "playing");
	g_object_class_override_property (object_class, PROP_PAUSABLE, "pausable");
}

static void
rb_dbus_proxy_init (RbDBusProxy *dproxy)
{
	GError *error = NULL;

	dproxy->priv = g_new0 (RbDBusProxyPrivate, 1);

	dproxy->priv->song = NULL;
	dproxy->priv->rating = 0.0;

	dproxy->priv->connected = FALSE;
	dproxy->priv->playing = FALSE;
	dproxy->priv->pausable = FALSE;
	dproxy->priv->uri = NULL;
	dproxy->priv->elapsed = -1;

	dproxy->priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (dproxy->priv->connection == NULL)
	{
		g_warning ("Failed to open connection to bus: %s", error->message);
		g_error_free (error);
		dproxy->priv->bus_proxy = NULL;
		dproxy->priv->player_proxy = NULL;
		dproxy->priv->shell_proxy = NULL;
		return;
	}

	dproxy->priv->bus_proxy = dbus_g_proxy_new_for_name (dproxy->priv->connection,
							     DBUS_SERVICE_DBUS,
							     DBUS_PATH_DBUS,
							     DBUS_INTERFACE_DBUS);
	dbus_g_proxy_add_signal (dproxy->priv->bus_proxy,
				 "NameOwnerChanged",
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_STRING,
				 G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (dproxy->priv->bus_proxy,
				     "NameOwnerChanged",
				     G_CALLBACK (name_owner_changed_cb),
				     dproxy,
				     NULL);

	dproxy->priv->player_proxy = dbus_g_proxy_new_for_name (dproxy->priv->connection,
								"org.gnome.Rhythmbox",
								"/org/gnome/Rhythmbox/Player",
								"org.gnome.Rhythmbox.Player");
	dbus_g_proxy_add_signal (dproxy->priv->player_proxy,
				 "playingChanged",
				 G_TYPE_BOOLEAN,
				 G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (dproxy->priv->player_proxy,
				     "playingChanged",
				     G_CALLBACK (playing_changed_cb),
				     dproxy,
				     NULL);

	dbus_g_proxy_add_signal (dproxy->priv->player_proxy,
				 "playingUriChanged",
				 G_TYPE_STRING,
				 G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (dproxy->priv->player_proxy,
				     "playingUriChanged",
				     G_CALLBACK (uri_changed_cb),
				     dproxy,
				     NULL);

	dbus_g_proxy_add_signal (dproxy->priv->player_proxy,
				 "elapsedChanged",
				 G_TYPE_UINT,
				 G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (dproxy->priv->player_proxy,
				     "elapsedChanged",
				     G_CALLBACK (elapsed_changed_cb),
				     dproxy,
				     NULL);

	dproxy->priv->shell_proxy = dbus_g_proxy_new_for_name (dproxy->priv->connection,
							       "org.gnome.Rhythmbox",
							       "/org/gnome/Rhythmbox/Shell",
							       "org.gnome.Rhythmbox.Shell");
}

static void
rb_dbus_proxy_dispose (GObject *object)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (object);

	if (dproxy->priv->bus_proxy != NULL)
		g_object_unref (dproxy->priv->bus_proxy);

	if (dproxy->priv->player_proxy != NULL)
		g_object_unref (dproxy->priv->player_proxy);

	if (dproxy->priv->shell_proxy != NULL)
		g_object_unref (dproxy->priv->shell_proxy);

	if (dproxy->priv->connection != NULL)
		dbus_g_connection_unref (dproxy->priv->connection);

	parent_class->dispose (object);
}

static void
rb_dbus_proxy_finalize (GObject *object)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (object);

	g_free (dproxy->priv);

	parent_class->finalize (object);
}

/* Public interface */

RbProxy *
rb_dbus_proxy_new (void)
{
	RbDBusProxy *dproxy;

	dproxy = g_object_new (RB_TYPE_DBUS_PROXY, NULL);

	refresh (dproxy);

	return RB_PROXY (dproxy);
}

/* GObject overrides */

static void
rb_dbus_proxy_get_property (GObject *object,
			    guint prop_id,
			    GValue *value,
			    GParamSpec *pspec)
{
	RbDBusProxy *dproxy;

	g_return_if_fail (object != NULL);

	dproxy = RB_DBUS_PROXY (object);

	switch (prop_id)
	{
	case PROP_SONG:
		g_value_set_object (value, dproxy->priv->song);
		break;
	case PROP_ELAPSED:
		g_value_set_long (value, (glong) dproxy->priv->elapsed);
		break;
	case PROP_RATING:
		g_value_set_double (value, dproxy->priv->rating);
		break;
	case PROP_CONNECTED:
		g_value_set_boolean (value, dproxy->priv->connected);
		break;
	case PROP_PLAYING:
		g_value_set_boolean (value, dproxy->priv->playing);
		break;
	case PROP_PAUSABLE:
		g_value_set_boolean (value, dproxy->priv->pausable);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
rb_dbus_proxy_set_property (GObject *object,
			    guint prop_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (object);

	g_return_if_fail (dproxy != NULL);

	switch (prop_id)
	{
	case PROP_RATING:
		rate_current_song (dproxy, g_value_get_double (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

/* RbProxy overrides */

static void
rb_dbus_proxy_play_pause_stop (RbProxy *proxy)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (proxy);

	g_return_if_fail (dproxy != NULL);

	if (dproxy->priv->player_proxy == NULL)
		return;

	dbus_g_proxy_call_no_reply (dproxy->priv->player_proxy,
				    "playPause",
				    G_TYPE_BOOLEAN,
				    FALSE,
				    G_TYPE_INVALID);
}

static void
rb_dbus_proxy_previous (RbProxy *proxy)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (proxy);

	g_return_if_fail (dproxy != NULL);

	if (dproxy->priv->player_proxy == NULL)
		return;

	dbus_g_proxy_call_no_reply (dproxy->priv->player_proxy,
				    "previous",
				    G_TYPE_INVALID);
}

static void
rb_dbus_proxy_next (RbProxy *proxy)
{
	RbDBusProxy *dproxy = RB_DBUS_PROXY (proxy);

	g_return_if_fail (dproxy != NULL);

	if (dproxy->priv->player_proxy == NULL)
		return;

	dbus_g_proxy_call_no_reply (dproxy->priv->player_proxy,
				    "next",
				    G_TYPE_INVALID);
}

/* Internal functions */

static void
rate_current_song (RbDBusProxy *dproxy, gdouble rating)
{
	GValue rating_value;

	if (dproxy->priv->shell_proxy == NULL)
		return;

	memset (&rating_value, 0, sizeof (rating_value));
	g_value_init (&rating_value, G_TYPE_DOUBLE);
	g_value_set_double (&rating_value, rating);

	dbus_g_proxy_call_no_reply (dproxy->priv->shell_proxy,
				    "setSongProperty",
				    G_TYPE_STRING, dproxy->priv->uri,
				    G_TYPE_STRING, "rating",
				    G_TYPE_VALUE, &rating_value,
				    G_TYPE_INVALID,
				    G_TYPE_INVALID);

	dproxy->priv->rating = rating;
	g_object_notify (G_OBJECT (dproxy), "rating");
}

static void
refresh (RbDBusProxy *dproxy)
{
	GError *error = NULL;

	dproxy->priv->connected = FALSE;
	g_free (dproxy->priv->uri);

	if (dproxy->priv->song != NULL)
	{
		g_object_unref (dproxy->priv->song);
		dproxy->priv->song = NULL;
		dproxy->priv->rating = 0.0;
		dproxy->priv->pausable = FALSE;
	}

	if (dproxy->priv->bus_proxy == NULL ||
	    !dbus_g_proxy_call (dproxy->priv->bus_proxy, "NameHasOwner", &error,
		    		G_TYPE_STRING, "org.gnome.Rhythmbox",
				G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &dproxy->priv->connected,
				G_TYPE_INVALID))
	{
		if (error != NULL)
		{
			g_warning("NameHasOwner failed: %s", error->message);
			g_error_free (error);
		}
	}

	refresh_playing (dproxy);
	refresh_elapsed (dproxy);
	refresh_playing_uri (dproxy);

	cache_song_metadata (dproxy);

	g_object_notify (G_OBJECT (dproxy), "playing");
	g_object_notify (G_OBJECT (dproxy), "elapsed");
	g_object_notify (G_OBJECT (dproxy), "connected");
}

static void
refresh_playing (RbDBusProxy *dproxy)
{
	GError *error = NULL;

	if (!dproxy->priv->connected ||
	    dproxy->priv->player_proxy == NULL ||
	    !dbus_g_proxy_call (dproxy->priv->player_proxy, "getPlaying", &error,
		    		G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &dproxy->priv->playing,
				G_TYPE_INVALID))
	{
		if (error != NULL)
		{
			g_warning ("getPlaying failed: %s", error->message);
			g_error_free (error);
		}

		dproxy->priv->connected = FALSE;
		dproxy->priv->playing = FALSE;
	}
}

static void
refresh_elapsed (RbDBusProxy *dproxy)
{
	glong elapsed;
	GError *error = NULL;

	if (!dproxy->priv->connected ||
	    dproxy->priv->player_proxy == NULL ||
	    !dbus_g_proxy_call (dproxy->priv->player_proxy, "getElapsed", &error,
		    		G_TYPE_INVALID,
				G_TYPE_UINT, &elapsed,
				G_TYPE_INVALID))
	{
		if (error != NULL)
		{
			g_warning ("getElapsed failed: %s", error->message);
			g_error_free (error);
		}

		dproxy->priv->connected = FALSE;
		dproxy->priv->elapsed = -1;
	}
	else
		dproxy->priv->elapsed = elapsed;
}

static void
refresh_playing_uri (RbDBusProxy *dproxy)
{
	GError *error = NULL;

	if (!dproxy->priv->connected ||
	    dproxy->priv->player_proxy == NULL ||
	    !dbus_g_proxy_call (dproxy->priv->player_proxy, "getPlayingUri", &error,
		    		G_TYPE_INVALID,
				G_TYPE_STRING, &dproxy->priv->uri,
				G_TYPE_INVALID))
	{
		if (error != NULL)
		{
			g_warning ("getPlayingUri failed: %s", error->message);
			g_error_free (error);
		}

		dproxy->priv->connected = FALSE;
		dproxy->priv->uri = NULL;
	}

	if (dproxy->priv->uri != NULL && *dproxy->priv->uri == '\0')
	{
		g_free (dproxy->priv->uri);
		dproxy->priv->uri = NULL;
	}
}

static void
cache_song_metadata (RbDBusProxy *dproxy)
{
	GHashTable *table = NULL;
	GError *error = NULL;

	if (dproxy->priv->song != NULL)
		g_object_unref (dproxy->priv->song);

	if (dproxy->priv->uri == NULL)
	{
		dproxy->priv->song = NULL;
		dproxy->priv->rating = 0.0;
		dproxy->priv->pausable = FALSE;
	}
	else if (dbus_g_proxy_call (dproxy->priv->shell_proxy, "getSongProperties", &error,
				    G_TYPE_STRING, dproxy->priv->uri,
				    G_TYPE_INVALID,
				    dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), &table,
				    G_TYPE_INVALID))
	{
		GValue *value;

		const gchar *title = "Unknown";
		const gchar *artist = "Unknown";
		const gchar *album = "Unknown";
		glong duration = 1;
		gdouble rating = 0.0;

		value = (GValue *) g_hash_table_lookup (table, "title");
		if (value != NULL && G_VALUE_HOLDS_STRING (value))
			title = g_value_get_string (value);

		value = (GValue *) g_hash_table_lookup (table, "artist");
		if (value != NULL && G_VALUE_HOLDS_STRING (value))
			artist = g_value_get_string (value);

		value = (GValue *) g_hash_table_lookup (table, "album");
		if (value != NULL && G_VALUE_HOLDS_STRING (value))
			album = g_value_get_string (value);

		value = (GValue *) g_hash_table_lookup (table, "duration");
		if (value != NULL && G_VALUE_HOLDS_UINT (value))
			duration = g_value_get_uint (value);

		value = (GValue *) g_hash_table_lookup (table, "rating");
		if (value != NULL && G_VALUE_HOLDS_DOUBLE (value))
			rating = g_value_get_double (value);

		dproxy->priv->song = rb_song_metadata_new (title, artist, album, duration);
		dproxy->priv->rating = rating;
		dproxy->priv->pausable = (duration > 0);

		g_hash_table_destroy (table);
	}
	else
	{
		g_error_free (error);
		dproxy->priv->song = NULL;
		dproxy->priv->rating = 0.0;
		dproxy->priv->pausable = FALSE;
	}

	g_object_notify (G_OBJECT (dproxy), "song");
	g_object_notify (G_OBJECT (dproxy), "rating");
	g_object_notify (G_OBJECT (dproxy), "pausable");
}

/* Callbacks */

static void
name_owner_changed_cb (DBusGProxy *bus_proxy,
		       const gchar *name,
		       const gchar *old_owner,
		       const gchar *new_owner,
		       RbDBusProxy *dproxy)
{
	if (strcmp (name, "org.gnome.Rhythmbox") == 0)
		refresh (dproxy);
}

static void
playing_changed_cb (DBusGProxy *player_proxy,
		    gboolean playing,
		    RbDBusProxy *dproxy)
{
	dproxy->priv->playing = playing;
	g_object_notify (G_OBJECT (dproxy), "playing");

	/**
	 * Since uri_changed_cb doesn't get called when playback finishes,
	 * we have to check for not-playing AND null-elapsed-time.
	 */

	if (dproxy->priv->elapsed == (guint) -1 && !playing &&
	    dproxy->priv->uri != NULL)
	{
		g_free (dproxy->priv->uri);
		dproxy->priv->uri = NULL;
		cache_song_metadata (dproxy);
	}
}

static void
uri_changed_cb (DBusGProxy *player_proxy,
		const gchar *uri,
		RbDBusProxy *dproxy)
{
	g_free (dproxy->priv->uri);
	dproxy->priv->uri = g_strdup (uri);

	cache_song_metadata (dproxy);
}

static void
elapsed_changed_cb (DBusGProxy *player_proxy,
		     guint elapsed,
		     RbDBusProxy *dproxy)
{
	dproxy->priv->elapsed = elapsed;

	/**
	 * Since uri_changed_cb doesn't get called when playback finishes,
	 * we have to check for not-playing AND null-elapsed-time.
	 */

	if (elapsed == (guint) -1 && !dproxy->priv->playing &&
	    dproxy->priv->uri != NULL)
	{
		g_free (dproxy->priv->uri);
		dproxy->priv->uri = NULL;
		cache_song_metadata (dproxy);
	}

	g_object_notify (G_OBJECT (dproxy), "elapsed");
}
