/*
 * Header of the immutable metadata of a song.
 *
 * Rhythmbox Applet
 * Copyright (C) 2004-2005 Paul Kuliniewicz <kuliniew@purdue.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.
 */

#include <config.h>

#include "rb-song-metadata.h"

struct _RbSongMetadataPrivate
{
	gchar *title;
	gchar *artist;
	gchar *album;
	glong duration;
};

typedef enum
{
	PROP_NONE,
	PROP_TITLE,
	PROP_ARTIST,
	PROP_ALBUM,
	PROP_DURATION,
} RbSongMetadataProperty;

static GObjectClass *parent_class;

/* Functional declarations */

static void rb_song_metadata_class_init (RbSongMetadataClass *klass);
static void rb_song_metadata_init (RbSongMetadata *song);

static void rb_song_metadata_finalize (GObject *object);
static void rb_song_metadata_get_property (GObject *object,
					   guint prop_id,
					   GValue *value,
					   GParamSpec *pspec);

/* GType stuff */

GType
rb_song_metadata_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (RbSongMetadataClass),			/* class_size */
			NULL,						/* base_init */
			NULL,						/* base_finalize */
			(GClassInitFunc) rb_song_metadata_class_init,	/* class_init */
			NULL,						/* class_finalize */
			NULL,						/* class_data */
			sizeof (RbSongMetadata),			/* instance_size */
			0,						/* n_preallocs */
			(GInstanceInitFunc) rb_song_metadata_init,	/* instance_init */
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "RbSongMetadata", &info, 0);
	}

	return type;
}

static void
rb_song_metadata_class_init (RbSongMetadataClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;
	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = rb_song_metadata_finalize;
	object_class->get_property = rb_song_metadata_get_property;

	g_object_class_install_property (object_class,
					 PROP_TITLE,
					 g_param_spec_string ("title",
						 	      "title",
							      "Title of the song",
							      "",
							      G_PARAM_READABLE));

	g_object_class_install_property (object_class,
					 PROP_ARTIST,
					 g_param_spec_string ("artist",
						 	      "artist",
							      "Artist of the song",
							      "",
							      G_PARAM_READABLE));

	g_object_class_install_property (object_class,
					 PROP_ALBUM,
					 g_param_spec_string ("album",
						 	      "album",
							      "Album of the song",
							      "",
							      G_PARAM_READABLE));

	g_object_class_install_property (object_class,
					 PROP_DURATION,
					 g_param_spec_long ("duration",
						 	    "duration",
							    "Duration of the song, in seconds",
							    -1,
							    G_MAXLONG,
							    -1,
							    G_PARAM_READABLE));
}

static void
rb_song_metadata_init (RbSongMetadata *song)
{
	song->priv = g_new0 (RbSongMetadataPrivate, 1);

	song->priv->title = NULL;
	song->priv->artist = NULL;
	song->priv->album = NULL;
	song->priv->duration = -1;
}

static void
rb_song_metadata_finalize (GObject *object)
{
	RbSongMetadata *song = RB_SONG_METADATA (object);

	g_free (song->priv->title);
	g_free (song->priv->artist);
	g_free (song->priv->album);

	g_free (song->priv);

	parent_class->finalize (object);
}

/* Public interface */

RbSongMetadata *
rb_song_metadata_new (const gchar *title,
		      const gchar *artist,
		      const gchar *album,
		      glong duration)
{
	RbSongMetadata *song;

	song = g_object_new (RB_TYPE_SONG_METADATA, NULL);
	song->priv->title = g_strdup (title);
	song->priv->artist = g_strdup (artist);
	song->priv->album = g_strdup (album);
	song->priv->duration = duration;

	return song;
}

/* GObject overrides */

static void
rb_song_metadata_get_property (GObject *object,
			       guint prop_id,
			       GValue *value,
			       GParamSpec *pspec)
{
	RbSongMetadata *song;

	g_return_if_fail (object != NULL);

	song = RB_SONG_METADATA (object);

	switch (prop_id)
	{
	case PROP_TITLE:
		g_value_set_string (value, song->priv->title);
		break;
	case PROP_ARTIST:
		g_value_set_string (value, song->priv->artist);
		break;
	case PROP_ALBUM:
		g_value_set_string (value, song->priv->album);
		break;
	case PROP_DURATION:
		g_value_set_long (value, song->priv->duration);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}
