/*
 * Implementation of the fancy tooltip for displaying song information.
 *
 * Rhythmbox Applet
 * Copyright (C) 2004-2006 Paul Kuliniewicz <kuliniew@purdue.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1301, USA.
 *
 */

#include <config.h>

#include "rb-song-tooltip.h"

#include "rb-song-metadata.h"

#include <glib/gi18n.h>
#include <gtk/gtklabel.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtkvbox.h>


struct RbSongTooltipPrivate
{
	RbProxy *proxy;
	gint song_minutes;
	gint song_seconds;

	GtkTooltips *tooltips;
	GtkWidget *box;
	GtkWidget *title;
	GtkWidget *whence;
	GtkWidget *time;
};

static GObjectClass *parent_class;

/* Function declarations */

static void rb_song_tooltip_class_init (RbSongTooltipClass *klass);
static void rb_song_tooltip_init (RbSongTooltip *tip);

static void rb_song_tooltip_dispose (GObject *object);
static void rb_song_tooltip_finalize (GObject *object);

static void song_changed_cb (RbProxy *proxy,
			     GParamSpec *pspec,
			     RbSongTooltip *tip);
static void time_updated_cb (RbProxy *proxy,
			     GParamSpec *pspec,
			     RbSongTooltip *tip);

/* GType stuff */

GType
rb_song_tooltip_get_type (void)
{
	static GType this_type = 0;

	if (!this_type)
	{
		static const GTypeInfo this_info = {
			sizeof (RbSongTooltipClass),
			NULL,
			NULL,
			(GClassInitFunc) rb_song_tooltip_class_init,
			NULL,
			NULL,
			sizeof (RbSongTooltip),
			0,
			(GInstanceInitFunc) rb_song_tooltip_init,
			NULL
		};
		this_type = g_type_register_static (G_TYPE_OBJECT,
						    "RbSongTooltip",
						    &this_info,
						    0);
	}

	return this_type;
}

static void
rb_song_tooltip_class_init (RbSongTooltipClass *klass)
{
	GObjectClass *object_class = (GObjectClass *) klass;

	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = rb_song_tooltip_dispose;
	object_class->finalize = rb_song_tooltip_finalize;
}

static void
rb_song_tooltip_init (RbSongTooltip *tip)
{
	tip->priv = g_new0 (RbSongTooltipPrivate, 1);

	tip->priv->proxy = NULL;
	tip->priv->song_minutes = 0;
	tip->priv->song_seconds = 0;

	tip->priv->tooltips = gtk_tooltips_new ();
	g_object_ref (tip->priv->tooltips);
	gtk_object_sink (GTK_OBJECT (tip->priv->tooltips));
	gtk_tooltips_force_window (tip->priv->tooltips);

	tip->priv->box = gtk_vbox_new (FALSE, 2);

	tip->priv->title = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (tip->priv->title), 0.0, 0.5);
	gtk_label_set_justify (GTK_LABEL (tip->priv->title),
			       GTK_JUSTIFY_LEFT);
	gtk_box_pack_start_defaults (GTK_BOX (tip->priv->box),
				     tip->priv->title);

	tip->priv->whence = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (tip->priv->whence), 0.0, 0.5);
	gtk_label_set_justify (GTK_LABEL (tip->priv->whence),
			       GTK_JUSTIFY_LEFT);
	gtk_box_pack_start_defaults (GTK_BOX (tip->priv->box),
				     tip->priv->whence);

	tip->priv->time = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (tip->priv->time), 0.0, 0.5);
	gtk_label_set_justify (GTK_LABEL (tip->priv->time),
			       GTK_JUSTIFY_LEFT);
	gtk_box_pack_start_defaults (GTK_BOX (tip->priv->box),
				     tip->priv->time);

	/* Such a horrible abuse of GtkTooltips... */
	g_object_ref (tip->priv->tooltips->tip_label);
	gtk_container_remove (GTK_CONTAINER (tip->priv->tooltips->tip_window),
			      tip->priv->tooltips->tip_label);
	gtk_container_add (GTK_CONTAINER (tip->priv->tooltips->tip_window),
			   tip->priv->box);

	gtk_widget_show (tip->priv->title);
	gtk_widget_show (tip->priv->box);
}

static void
rb_song_tooltip_dispose (GObject *object)
{
	RbSongTooltip *tip = RB_SONG_TOOLTIP (object);

	g_signal_handlers_disconnect_by_func (tip->priv->proxy, song_changed_cb, tip);
	g_signal_handlers_disconnect_by_func (tip->priv->proxy, time_updated_cb, tip);

	g_object_unref (tip->priv->tooltips->tip_label);
	g_object_unref (tip->priv->tooltips);
	g_object_unref (tip->priv->proxy);

	parent_class->dispose (object);
}

static void
rb_song_tooltip_finalize (GObject *object)
{
	RbSongTooltip *tip = RB_SONG_TOOLTIP (object);

	parent_class->finalize (object);
}

RbSongTooltip *
rb_song_tooltip_new (RbProxy *proxy)
{
	RbSongTooltip *tip;

	g_return_val_if_fail (proxy != NULL, NULL);

	tip = g_object_new (RB_TYPE_SONG_TOOLTIP, NULL);
	tip->priv->proxy = proxy;
	g_object_ref (proxy);

	g_signal_connect (proxy, "notify::song",
			  G_CALLBACK (song_changed_cb), tip);
	g_signal_connect (proxy, "notify::elapsed",
			  G_CALLBACK (time_updated_cb), tip);

	/* Abuse the event handler to initialize the display. */
	song_changed_cb (tip->priv->proxy, NULL, tip);

	return tip;
}

/* Interface functions */

void
rb_song_tooltip_attach (RbSongTooltip *tip, GtkWidget *owner)
{
	gtk_tooltips_set_tip (tip->priv->tooltips, owner, "", "");
}

/* Callbacks */

static void
song_changed_cb (RbProxy *proxy,
		 GParamSpec *pspec,
		 RbSongTooltip *tip)
{
	RbSongMetadata *song;

	g_object_get (proxy, "song", &song, NULL);

	if (song != NULL)
	{
		gchar *text;

		gchar *title;
		gchar *artist;
		gchar *album;
		glong duration;

		g_object_get (song,
			      "title", &title,
			      "artist", &artist,
			      "album", &album,
			      "duration", &duration,
			      NULL);

		text = g_markup_printf_escaped ("<big><b>%s</b></big>",
						title);
		gtk_label_set_markup (GTK_LABEL (tip->priv->title), text);
		g_free (text);

		if (duration > 0)
		{
			/* Playing an ordinary file. */
			text = g_markup_printf_escaped (
				/* To translator: from <i>ALBUM</i> by <i>ARTIST</i> */
				_("from <i>%s</i> by <i>%s</i>"),
				album, artist);
			gtk_label_set_markup (GTK_LABEL (tip->priv->whence),
					      text);
			g_free (text);

			gtk_widget_show (tip->priv->whence);
		}
		else
		{
			/* Playing an Internet radio station. */
			gtk_widget_hide (tip->priv->whence);
		}

		tip->priv->song_minutes = duration / 60;
		tip->priv->song_seconds = duration % 60;

		gtk_widget_show (tip->priv->time);

		time_updated_cb (proxy, NULL, tip);

		g_free(title);
		g_free(artist);
		g_free(album);

		g_object_unref (G_OBJECT (song));
	}
	else
	{
		gtk_label_set_text (GTK_LABEL (tip->priv->title),
				    _("Not playing"));
		gtk_widget_hide (tip->priv->whence);
		gtk_widget_hide (tip->priv->time);
	}
}

static void
time_updated_cb (RbProxy *proxy, GParamSpec *pspec, RbSongTooltip *tip)
{
	glong play_time;

	g_object_get (proxy, "elapsed", &play_time, NULL);

	if (play_time >= 0)
	{
		gint minutes;
		gint seconds;
		gchar *text;

		minutes = (int) (play_time / 60);
		seconds = (int) (play_time % 60);

		if (tip->priv->song_minutes > 0 || tip->priv->song_seconds > 0)
		{
			/* Playing an ordinary file. */
			/* To translator: CURRENT-TIME of TOTAL-TIME */
			text = g_strdup_printf (_("%d:%.2d of %d:%.2d"),
						minutes,
						seconds,
						tip->priv->song_minutes,
						tip->priv->song_seconds);
		}
		else
		{
			/* Playing an Internet radio station. */
			text = g_strdup_printf("%d:%.2d",
					       minutes,
					       seconds);
		}

		gtk_label_set_text (GTK_LABEL (tip->priv->time), text);
		g_free (text);
	}
}
